@@ -171,6 +171,13 @@ func originalExecNode(n *node, exec bltn) *node {
 	return originalNode
 }
 
+// cloned from net/http/server.go , so we can enforce a similar behavior:
+// in the stdlib, this error is used as sentinel in panic triggered e.g. on
+// request cancellation, in order to catch it and suppress it in a following defer.
+// in yaegi, we use it to suppress a "panic" log message that happens in the
+// same circumstances.
+var errAbortHandler = errors.New("net/http: abort Handler")
+
 // Functions set to run during execution of CFG.
 
 // runCfg executes a node AST by walking its CFG and running node builtin at each step.
@@ -187,7 +194,13 @@ func runCfg(n *node, f *frame, funcNode, callNode *node) {
 			if oNode == nil {
 				oNode = n
 			}
-			fmt.Fprintln(n.interp.stderr, oNode.cfgErrorf("panic"))
+			errorer, ok := f.recovered.(error)
+			// in this specific case, the stdlib would/will suppress the panic, so we
+			// suppress the logging here accordingly, to get a similar and consistent
+			// behavior.
+			if !ok || errorer.Error() != errAbortHandler.Error() {
+				fmt.Fprintln(n.interp.stderr, oNode.cfgErrorf("panic"))
+			}
 			f.mutex.Unlock()
 			panic(f.recovered)
 		}
@@ -401,7 +414,7 @@ func typeAssert(n *node, withResult, withOk bool) {
 			ok = v.IsValid()
 			if !ok {
 				if !withOk {
-					panic(fmt.Sprintf("interface conversion: interface {} is nil, not %s", rtype.String()))
+					panic(n.cfgErrorf("interface conversion: interface {} is nil, not %s", rtype.String()))
 				}
 				return next
 			}
@@ -409,7 +422,7 @@ func typeAssert(n *node, withResult, withOk bool) {
 			if !ok {
 				if !withOk {
 					method := firstMissingMethod(leftType, rtype)
-					panic(fmt.Sprintf("interface conversion: %s is not %s: missing method %s", leftType.String(), rtype.String(), method))
+					panic(n.cfgErrorf("interface conversion: %s is not %s: missing method %s", leftType.String(), rtype.String(), method))
 				}
 				return next
 			}
@@ -430,14 +443,18 @@ func typeAssert(n *node, withResult, withOk bool) {
 			concrete := val.Interface()
 			ctyp := reflect.TypeOf(concrete)
 
+			if vv, ok := concrete.(valueInterface); ok {
+				ctyp = vv.value.Type()
+				concrete = vv.value.Interface()
+			}
 			ok = canAssertTypes(ctyp, rtype)
 			if !ok {
 				if !withOk {
 					// TODO(mpl): think about whether this should ever happen.
 					if ctyp == nil {
-						panic(fmt.Sprintf("interface conversion: interface {} is nil, not %s", rtype.String()))
+						panic(n.cfgErrorf("interface conversion: interface {} is nil, not %s", rtype.String()))
 					}
-					panic(fmt.Sprintf("interface conversion: interface {} is %s, not %s", ctyp.String(), rtype.String()))
+					panic(n.cfgErrorf("interface conversion: interface {} is %s, not %s", ctyp.String(), rtype.String()))
 				}
 				return next
 			}
@@ -462,7 +479,7 @@ func typeAssert(n *node, withResult, withOk bool) {
 			}
 			if !ok {
 				if !withOk {
-					panic(fmt.Sprintf("interface conversion: interface {} is nil, not %s", rtype.String()))
+					panic(n.cfgErrorf("interface conversion: interface {} is nil, not %s", rtype.String()))
 				}
 				return next
 			}
@@ -475,7 +492,7 @@ func typeAssert(n *node, withResult, withOk bool) {
 			if !ok {
 				if !withOk {
 					method := firstMissingMethod(v.Type(), rtype)
-					panic(fmt.Sprintf("interface conversion: %s is not %s: missing method %s", v.Type().String(), rtype.String(), method))
+					panic(n.cfgErrorf("interface conversion: %s is not %s: missing method %s", v.Type().String(), rtype.String(), method))
 				}
 				return next
 			}
@@ -495,7 +512,7 @@ func typeAssert(n *node, withResult, withOk bool) {
 			if !ok || !v.value.IsValid() {
 				ok = false
 				if !withOk {
-					panic(fmt.Sprintf("interface conversion: interface {} is nil, not %s", rtype.String()))
+					panic(n.cfgErrorf("interface conversion: interface {} is nil, not %s", rtype.String()))
 				}
 				return next
 			}
@@ -503,7 +520,7 @@ func typeAssert(n *node, withResult, withOk bool) {
 			ok = canAssertTypes(v.value.Type(), rtype)
 			if !ok {
 				if !withOk {
-					panic(fmt.Sprintf("interface conversion: interface {} is %s, not %s", v.value.Type().String(), rtype.String()))
+					panic(n.cfgErrorf("interface conversion: interface {} is %s, not %s", v.value.Type().String(), rtype.String()))
 				}
 				return next
 			}
@@ -562,29 +579,11 @@ func convert(n *node) {
 		return
 	}
 
-	if isFuncSrc(n.child[0].typ) && isFuncSrc(c.typ) {
-		value := genValue(c)
-		n.exec = func(f *frame) bltn {
-			n, ok := value(f).Interface().(*node)
-			if !ok || !n.typ.convertibleTo(c.typ) {
-				panic("cannot convert")
-			}
-			n1 := *n
-			n1.typ = c.typ
-			dest(f).Set(reflect.ValueOf(&n1))
-			return next
-		}
-		return
-	}
-
 	doConvert := true
 	var value func(*frame) reflect.Value
 	switch {
 	case isFuncSrc(c.typ):
 		value = genFunctionWrapper(c)
-	case isFuncSrc(n.child[0].typ) && c.typ.cat == valueT:
-		doConvert = false
-		value = genValueNode(c)
 	default:
 		value = genValue(c)
 	}
@@ -655,8 +654,8 @@ func assign(n *node) {
 
 	for i := 0; i < n.nleft; i++ {
 		dest, src := n.child[i], n.child[sbase+i]
-		if isFuncSrc(src.typ) && isField(dest) {
-			svalue[i] = genFunctionWrapper(src)
+		if isNamedFuncSrc(src.typ) {
+			svalue[i] = genFuncValue(src)
 		} else {
 			svalue[i] = genDestValue(dest.typ, src)
 		}
@@ -710,8 +709,6 @@ func assign(n *node) {
 	for i := range types {
 		var t reflect.Type
 		switch typ := n.child[sbase+i].typ; {
-		case isFuncSrc(typ):
-			t = reflect.TypeOf((*node)(nil))
 		case isInterfaceSrc(typ):
 			t = valueInterfaceType
 		default:
@@ -948,9 +945,6 @@ func genFunctionWrapper(n *node) func(*frame) reflect.Value {
 	var def *node
 	var ok bool
 
-	if n.kind == basicLit {
-		return func(f *frame) reflect.Value { return n.rval }
-	}
 	if def, ok = n.val.(*node); !ok {
 		return genValueAsFunctionWrapper(n)
 	}
@@ -959,11 +953,7 @@ func genFunctionWrapper(n *node) func(*frame) reflect.Value {
 	var rcvr func(*frame) reflect.Value
 
 	if n.recv != nil {
-		if n.recv.node.typ.cat != defRecvType(def).cat {
-			rcvr = genValueRecvIndirect(n)
-		} else {
-			rcvr = genValueRecv(n)
-		}
+		rcvr = genValueRecv(n)
 	}
 	funcType := n.typ.TypeOf()
 
@@ -985,15 +975,20 @@ func genFunctionWrapper(n *node) func(*frame) reflect.Value {
 				// Copy method receiver as first argument.
 				src, dest := rcvr(f), d[numRet]
 				sk, dk := src.Kind(), dest.Kind()
+				for {
+					vs, ok := src.Interface().(valueInterface)
+					if !ok {
+						break
+					}
+					src = vs.value
+					sk = src.Kind()
+				}
 				switch {
 				case sk == reflect.Ptr && dk != reflect.Ptr:
 					dest.Set(src.Elem())
 				case sk != reflect.Ptr && dk == reflect.Ptr:
 					dest.Set(src.Addr())
 				default:
-					if wrappedSrc, ok := src.Interface().(valueInterface); ok {
-						src = wrappedSrc.value
-					}
 					dest.Set(src)
 				}
 				d = d[numRet+1:]
@@ -1011,8 +1006,6 @@ func genFunctionWrapper(n *node) func(*frame) reflect.Value {
 					d[i].Set(arg)
 				case isInterfaceSrc(typ):
 					d[i].Set(reflect.ValueOf(valueInterface{value: arg.Elem()}))
-				case isFuncSrc(typ) && arg.Kind() == reflect.Func:
-					d[i].Set(reflect.ValueOf(genFunctionNode(arg)))
 				default:
 					d[i].Set(arg)
 				}
@@ -1021,21 +1014,11 @@ func genFunctionWrapper(n *node) func(*frame) reflect.Value {
 			// Interpreter code execution.
 			runCfg(start, fr, def, n)
 
-			result := fr.data[:numRet]
-			for i, r := range result {
-				if v, ok := r.Interface().(*node); ok {
-					result[i] = genFunctionWrapper(v)(f)
-				}
-			}
-			return result
+			return fr.data[:numRet]
 		})
 	}
 }
 
-func genFunctionNode(v reflect.Value) *node {
-	return &node{kind: funcType, action: aNop, rval: v, typ: valueTOf(v.Type())}
-}
-
 func genInterfaceWrapper(n *node, typ reflect.Type) func(*frame) reflect.Value {
 	value := genValue(n)
 	if typ == nil || typ.Kind() != reflect.Interface || typ.NumMethod() == 0 || n.typ.cat == valueT {
@@ -1174,11 +1157,6 @@ func call(n *node) {
 		// Compute method receiver value.
 		values = append(values, genValueRecv(c0))
 		method = true
-	case len(c0.child) > 0 && c0.child[0].typ != nil && isInterfaceSrc(c0.child[0].typ):
-		recvIndexLater = true
-		values = append(values, genValueBinRecv(c0, &receiver{node: c0.child[0]}))
-		value = genValueBinMethodOnInterface(n, value)
-		method = true
 	case c0.action == aMethod:
 		// Add a place holder for interface method receiver.
 		values = append(values, nil)
@@ -1240,7 +1218,7 @@ func call(n *node) {
 			case isInterfaceBin(arg):
 				values = append(values, genInterfaceWrapper(c, arg.rtype))
 			case isFuncSrc(arg):
-				values = append(values, genValueNode(c))
+				values = append(values, genFuncValue(c))
 			default:
 				values = append(values, genValue(c))
 			}
@@ -1305,21 +1283,47 @@ func call(n *node) {
 		var ok bool
 
 		bf := value(f)
+
 		if def, ok = bf.Interface().(*node); ok {
 			bf = def.rval
 		}
 
 		// Call bin func if defined
 		if bf.IsValid() {
+			var callf func([]reflect.Value) []reflect.Value
+
+			// Lambda definitions are necessary here. Due to reflect internals,
+			// having `callf = bf.Call` or `callf = bf.CallSlice` does not work.
+			//nolint:gocritic
+			if hasVariadicArgs {
+				callf = func(in []reflect.Value) []reflect.Value { return bf.CallSlice(in) }
+			} else {
+				callf = func(in []reflect.Value) []reflect.Value { return bf.Call(in) }
+			}
+
+			if method && len(values) > bf.Type().NumIn() {
+				// The receiver is already passed in the function wrapper, skip it.
+				values = values[1:]
+			}
+
+			if goroutine {
+				// Goroutine's arguments should be copied.
+				in := make([]reflect.Value, len(values))
+				for i, v := range values {
+					value := v(f)
+					in[i] = reflect.New(value.Type()).Elem()
+					in[i].Set(value)
+				}
+
+				go callf(in)
+				return tnext
+			}
+
 			in := make([]reflect.Value, len(values))
 			for i, v := range values {
 				in[i] = v(f)
 			}
-			if goroutine {
-				go bf.Call(in)
-				return tnext
-			}
-			out := bf.Call(in)
+			out := callf(in)
 			for i, v := range rvalues {
 				if v != nil {
 					v(f).Set(out[i])
@@ -1553,12 +1557,12 @@ func callBin(n *node) {
 			}
 
 			switch {
-			case isFuncSrc(c.typ):
-				values = append(values, genFunctionWrapper(c))
 			case isEmptyInterface(c.typ):
 				values = append(values, genValue(c))
 			case isInterfaceSrc(c.typ):
 				values = append(values, genValueInterfaceValue(c))
+			case isFuncSrc(c.typ):
+				values = append(values, genFunctionWrapper(c))
 			case c.typ.cat == arrayT || c.typ.cat == variadicT:
 				if isEmptyInterface(c.typ.val) {
 					values = append(values, genValueArray(c))
@@ -1906,8 +1910,10 @@ func getIndexMap2(n *node) {
 
 const fork = true // Duplicate frame in frame.clone().
 
+// getFunc compiles a closure function generator for anonymous functions.
 func getFunc(n *node) {
-	dest := genValue(n)
+	i := n.findex
+	l := n.level
 	next := getExec(n.tnext)
 
 	n.exec = func(f *frame) bltn {
@@ -1915,7 +1921,42 @@ func getFunc(n *node) {
 		nod := *n
 		nod.val = &nod
 		nod.frame = fr
-		dest(f).Set(reflect.ValueOf(&nod))
+		def := &nod
+		numRet := len(def.typ.ret)
+
+		fct := reflect.MakeFunc(nod.typ.TypeOf(), func(in []reflect.Value) []reflect.Value {
+			// Allocate and init local frame. All values to be settable and addressable.
+			fr2 := newFrame(fr, len(def.types), fr.runid())
+			d := fr2.data
+			for i, t := range def.types {
+				d[i] = reflect.New(t).Elem()
+			}
+			d = d[numRet:]
+
+			// Copy function input arguments in local frame.
+			for i, arg := range in {
+				if i >= len(d) {
+					// In case of unused arg, there may be not even a frame entry allocated, just skip.
+					break
+				}
+				typ := def.typ.arg[i]
+				switch {
+				case isEmptyInterface(typ) || typ.TypeOf() == valueInterfaceType:
+					d[i].Set(arg)
+				case isInterfaceSrc(typ):
+					d[i].Set(reflect.ValueOf(valueInterface{value: arg.Elem()}))
+				default:
+					d[i].Set(arg)
+				}
+			}
+
+			// Interpreter code execution.
+			runCfg(def.child[3].start, fr2, def, n)
+
+			return fr2.data[:numRet]
+		})
+
+		getFrame(f, l).data[i] = fct
 		return next
 	}
 }
@@ -1931,7 +1972,7 @@ func getMethod(n *node) {
 		nod.val = &nod
 		nod.recv = n.recv
 		nod.frame = fr
-		getFrame(f, l).data[i] = reflect.ValueOf(&nod)
+		getFrame(f, l).data[i] = genFuncValue(&nod)(f)
 		return next
 	}
 }
@@ -1998,7 +2039,7 @@ func getMethodByName(n *node) {
 		nod.val = &nod
 		nod.recv = &receiver{nil, val.value, li}
 		nod.frame = fr
-		getFrame(f, l).data[i] = reflect.ValueOf(&nod)
+		getFrame(f, l).data[i] = genFuncValue(&nod)(f)
 		return next
 	}
 }
@@ -2380,27 +2421,12 @@ func _return(n *node) {
 		switch t := def.typ.ret[i]; t.cat {
 		case errorT:
 			values[i] = genInterfaceWrapper(c, t.TypeOf())
-		case aliasT:
-			if isInterfaceSrc(t) {
-				values[i] = genValueInterface(c)
-			} else {
-				values[i] = genValue(c)
-			}
 		case funcT:
 			values[i] = genValue(c)
-		case interfaceT:
-			if len(t.field) == 0 {
-				// empty interface case.
-				// we can't let genValueInterface deal with it, because we call on c,
-				// not on n, which means that the interfaceT knowledge is lost.
-				values[i] = genValue(c)
-				break
-			}
-			values[i] = genValueInterface(c)
 		case valueT:
 			switch t.rtype.Kind() {
 			case reflect.Interface:
-				values[i] = genInterfaceWrapper(c, t.rtype)
+				values[i] = genInterfaceWrapper(c, t.TypeOf())
 				continue
 			case reflect.Func:
 				values[i] = genFunctionWrapper(c)
@@ -2408,9 +2434,19 @@ func _return(n *node) {
 			}
 			fallthrough
 		default:
-			if c.typ.untyped {
-				values[i] = genValueAs(c, def.typ.ret[i].TypeOf())
-			} else {
+			switch {
+			case isInterfaceSrc(t):
+				if len(t.field) == 0 {
+					// empty interface case.
+					// we can't let genValueInterface deal with it, because we call on c,
+					// not on n, which means that the interfaceT knowledge is lost.
+					values[i] = genValue(c)
+					break
+				}
+				values[i] = genValueInterface(c)
+			case c.typ.untyped:
+				values[i] = genValueAs(c, t.TypeOf())
+			default:
 				values[i] = genValue(c)
 			}
 		}
@@ -2605,7 +2641,7 @@ func doCompositeBinStruct(n *node, hasType bool) {
 	next := getExec(n.tnext)
 	value := valueGenerator(n, n.findex)
 	typ := n.typ.rtype
-	if n.typ.cat == ptrT || n.typ.cat == aliasT {
+	if n.typ.cat == ptrT || n.typ.cat == linkedT {
 		typ = n.typ.val.rtype
 	}
 	child := n.child
@@ -2672,7 +2708,7 @@ func doComposite(n *node, hasType bool, keyed bool) {
 	value := valueGenerator(n, n.findex)
 	next := getExec(n.tnext)
 	typ := n.typ
-	if typ.cat == ptrT || typ.cat == aliasT {
+	if typ.cat == ptrT || typ.cat == linkedT {
 		typ = typ.val
 	}
 	child := n.child
@@ -2698,7 +2734,7 @@ func doComposite(n *node, hasType bool, keyed bool) {
 		switch {
 		case val.typ.cat == nilT:
 			values[fieldIndex] = func(*frame) reflect.Value { return reflect.New(rft).Elem() }
-		case isFuncSrc(val.typ):
+		case isNamedFuncSrc(val.typ):
 			values[fieldIndex] = genValueAsFunctionWrapper(val)
 		case isInterfaceSrc(ft) && (!isEmptyInterface(ft) || len(val.typ.method) > 0):
 			values[fieldIndex] = genValueInterface(val)
@@ -2996,15 +3032,17 @@ func _case(n *node) {
 					}
 					return fnext
 				}
-				vi := v.Interface().(valueInterface)
-				if vi.node == nil {
-					if typ.cat == nilT {
-						return tnext
+				if vi, ok := v.Interface().(valueInterface); ok {
+					if vi.node != nil {
+						if vi.node.typ.id() == typ.id() {
+							destValue(f).Set(vi.value)
+							return tnext
+						}
 					}
 					return fnext
 				}
-				if vi.node.typ.id() == typ.id() {
-					destValue(f).Set(vi.value)
+				if v.Type() == typ.TypeOf() {
+					destValue(f).Set(v)
 					return tnext
 				}
 				return fnext
@@ -3039,13 +3077,23 @@ func _case(n *node) {
 					}
 					return fnext
 				}
-				if v := val.Interface().(valueInterface).node; v != nil {
-					for _, typ := range types {
-						if v.typ.id() == typ.id() {
-							destValue(f).Set(val)
-							return tnext
+				if vi, ok := val.Interface().(valueInterface); ok {
+					if v := vi.node; v != nil {
+						for _, typ := range types {
+							if v.typ.id() == typ.id() {
+								destValue(f).Set(val)
+								return tnext
+							}
 						}
 					}
+					return fnext
+				}
+				vt := val.Type()
+				for _, typ := range types {
+					if vt == typ.TypeOf() {
+						destValue(f).Set(val)
+						return tnext
+					}
 				}
 				return fnext
 			}
@@ -3611,7 +3659,7 @@ func convertConstantValue(n *node) {
 	case constant.Int:
 		i, x := constant.Int64Val(c)
 		if !x {
-			panic(fmt.Sprintf("constant %s overflows int64", c.ExactString()))
+			panic(n.cfgErrorf("constant %s overflows int64", c.ExactString()))
 		}
 		v = reflect.ValueOf(int(i))
 	case constant.Float:
@@ -3841,11 +3889,7 @@ func slice0(n *node) {
 func isNil(n *node) {
 	var value func(*frame) reflect.Value
 	c0 := n.child[0]
-	if isFuncSrc(c0.typ) {
-		value = genValueAsFunctionWrapper(c0)
-	} else {
-		value = genValue(c0)
-	}
+	value = genValue(c0)
 	typ := n.typ.concrete().TypeOf()
 	isInterface := n.typ.TypeOf().Kind() == reflect.Interface
 	tnext := getExec(n.tnext)
@@ -3930,11 +3974,7 @@ func isNil(n *node) {
 func isNotNil(n *node) {
 	var value func(*frame) reflect.Value
 	c0 := n.child[0]
-	if isFuncSrc(c0.typ) {
-		value = genValueAsFunctionWrapper(c0)
-	} else {
-		value = genValue(c0)
-	}
+	value = genValue(c0)
 	typ := n.typ.concrete().TypeOf()
 	isInterface := n.typ.TypeOf().Kind() == reflect.Interface
 	tnext := getExec(n.tnext)
