@@ -17,7 +17,6 @@ type tcat uint
 // Types for go language.
 const (
 	nilT tcat = iota
-	aliasT
 	arrayT
 	binT
 	binPkgT
@@ -41,6 +40,7 @@ const (
 	int16T
 	int32T
 	int64T
+	linkedT
 	mapT
 	ptrT
 	sliceT
@@ -60,7 +60,6 @@ const (
 
 var cats = [...]string{
 	nilT:        "nilT",
-	aliasT:      "aliasT",
 	arrayT:      "arrayT",
 	binT:        "binT",
 	binPkgT:     "binPkgT",
@@ -75,14 +74,15 @@ var cats = [...]string{
 	float32T:    "float32",
 	float64T:    "float64T",
 	funcT:       "funcT",
+	genericT:    "genericT",
 	interfaceT:  "interfaceT",
 	intT:        "intT",
 	int8T:       "int8T",
 	int16T:      "int16T",
 	int32T:      "int32T",
 	int64T:      "int64T",
+	linkedT:     "linkedT",
 	mapT:        "mapT",
-	genericT:    "genericT",
 	ptrT:        "ptrT",
 	sliceT:      "sliceT",
 	srcPkgT:     "srcPkgT",
@@ -118,7 +118,7 @@ type itype struct {
 	cat          tcat          // Type category
 	field        []structField // Array of struct fields if structT or interfaceT
 	key          *itype        // Type of key element if MapT or nil
-	val          *itype        // Type of value element if chanT, chanSendT, chanRecvT, mapT, ptrT, aliasT, arrayT, sliceT, variadicT or genericT
+	val          *itype        // Type of value element if chanT, chanSendT, chanRecvT, mapT, ptrT, linkedT, arrayT, sliceT, variadicT or genericT
 	recv         *itype        // Receiver type for funcT or nil
 	arg          []*itype      // Argument types if funcT or nil
 	ret          []*itype      // Return types if funcT or nil
@@ -126,6 +126,7 @@ type itype struct {
 	method       []*node       // Associated methods or nil
 	constraint   []*itype      // For interfaceT: list of types part of interface set
 	ulconstraint []*itype      // For interfaceT: list of underlying types part of interface set
+	instance     []*itype      // For genericT: list of instantiated types
 	name         string        // name of type within its package for a defined type
 	path         string        // for a defined type, the package import path
 	length       int           // length of array if ArrayT
@@ -248,7 +249,7 @@ func namedOf(val *itype, path, name string, opts ...itypeOption) *itype {
 	if path != "" {
 		str = path + "." + name
 	}
-	t := &itype{cat: aliasT, val: val, path: path, name: name, str: str}
+	t := &itype{cat: linkedT, val: val, path: path, name: name, str: str}
 	for _, opt := range opts {
 		opt(t)
 	}
@@ -373,8 +374,8 @@ func structOf(t *itype, fields []structField, opts ...itypeOption) *itype {
 }
 
 // genericOf returns a generic type.
-func genericOf(val *itype, name string, opts ...itypeOption) *itype {
-	t := &itype{cat: genericT, name: name, str: name, val: val}
+func genericOf(val *itype, name, path string, opts ...itypeOption) *itype {
+	t := &itype{cat: genericT, name: name, path: path, str: name, val: val}
 	for _, opt := range opts {
 		opt(t)
 	}
@@ -446,12 +447,19 @@ func nodeType2(interp *Interpreter, sc *scope, n *node, seen []*node) (t *itype,
 		)
 		switch v := c0.rval; {
 		case v.IsValid():
-			// Size if defined by a constant litteral value.
+			// Size if defined by a constant literal value.
 			if isConstantValue(v.Type()) {
 				c := v.Interface().(constant.Value)
 				length = constToInt(c)
 			} else {
-				length = int(v.Int())
+				switch v.Type().Kind() {
+				case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
+					length = int(v.Int())
+				case reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64, reflect.Uintptr:
+					length = int(v.Uint())
+				default:
+					return nil, c0.cfgErrorf("non integer constant %v", v)
+				}
 			}
 		case c0.kind == ellipsisExpr:
 			// [...]T expression, get size from the length of composite array.
@@ -476,7 +484,11 @@ func nodeType2(interp *Interpreter, sc *scope, n *node, seen []*node) (t *itype,
 			length = int(vInt(sym.rval))
 		default:
 			// Size is defined by a numeric constant expression.
-			if _, err = interp.cfg(c0, sc, sc.pkgID, sc.pkgName); err != nil {
+			if _, err := interp.cfg(c0, sc, sc.pkgID, sc.pkgName); err != nil {
+				if strings.Contains(err.Error(), " undefined: ") {
+					incomplete = true
+					break
+				}
 				return nil, err
 			}
 			v, ok := c0.rval.Interface().(constant.Value)
@@ -587,7 +599,10 @@ func nodeType2(interp *Interpreter, sc *scope, n *node, seen []*node) (t *itype,
 		}
 
 		if isInterfaceSrc(dt) {
-			dt.val = t
+			// Set a new interface type preserving the concrete type (.val field).
+			t2 := *dt
+			t2.val = t
+			dt = &t2
 		}
 		t = dt
 
@@ -768,11 +783,16 @@ func nodeType2(interp *Interpreter, sc *scope, n *node, seen []*node) (t *itype,
 			}
 		}
 		if sym.kind == varTypeSym {
-			t = genericOf(sym.typ, n.ident, withNode(n), withScope(sc))
+			t = genericOf(sym.typ, n.ident, sc.pkgName, withNode(n), withScope(sc))
 		} else {
 			t = sym.typ
 		}
-		if t.incomplete && t.cat == aliasT && t.val != nil && t.val.cat != nilT {
+		if t == nil {
+			if t, err = nodeType2(interp, sc, sym.node, seen); err != nil {
+				return nil, err
+			}
+		}
+		if t.incomplete && t.cat == linkedT && t.val != nil && t.val.cat != nilT {
 			t.incomplete = false
 		}
 		if t.incomplete && t.node != n {
@@ -793,7 +813,11 @@ func nodeType2(interp *Interpreter, sc *scope, n *node, seen []*node) (t *itype,
 			return nil, err
 		}
 		if lt.incomplete {
-			t.incomplete = true
+			if t == nil {
+				t = lt
+			} else {
+				t.incomplete = true
+			}
 			break
 		}
 		switch lt.cat {
@@ -814,46 +838,59 @@ func nodeType2(interp *Interpreter, sc *scope, n *node, seen []*node) (t *itype,
 				break
 			}
 			// A generic type is being instantiated. Generate it.
-			g, err := genAST(sc, lt.node.anc, []*node{t1.node})
+			t, err = genType(interp, sc, name, lt, []*itype{t1}, seen)
 			if err != nil {
 				return nil, err
 			}
-			t, err = nodeType2(interp, sc, g.lastChild(), seen)
-			if err != nil {
-				return nil, err
-			}
-			sc.sym[name] = &symbol{index: -1, kind: typeSym, typ: t, node: g}
-
-			// Instantiate type methods (if any).
-			var pt *itype
-			if len(lt.method) > 0 {
-				pt = ptrOf(t, withNode(g), withScope(sc))
-			}
-			for _, nod := range lt.method {
-				gm, err := genAST(sc, nod, []*node{t1.node})
-				if err != nil {
-					return nil, err
-				}
-				if gm.typ, err = nodeType(interp, sc, gm.child[2]); err != nil {
-					return nil, err
-				}
-				t.addMethod(gm)
-				if rtn := gm.child[0].child[0].lastChild(); rtn.kind == starExpr {
-					// The receiver is a pointer on a generic type.
-					pt.addMethod(gm)
-					rtn.typ = pt
-				}
-				// Compile method CFG.
-				if _, err = interp.cfg(gm, sc, sc.pkgID, sc.pkgName); err != nil {
-					return nil, err
-				}
-				// Generate closures for function body.
-				if err = genRun(gm); err != nil {
-					return nil, err
-				}
-			}
 		}
 
+	case indexListExpr:
+		// Similar to above indexExpr for generic types, but handle multiple type parameters.
+		var lt *itype
+		if lt, err = nodeType2(interp, sc, n.child[0], seen); err != nil {
+			return nil, err
+		}
+		if lt.incomplete {
+			if t == nil {
+				t = lt
+			} else {
+				t.incomplete = true
+			}
+			break
+		}
+
+		// Index list expressions can be used only in context of generic types.
+		if lt.cat != genericT {
+			err = n.cfgErrorf("not a generic type: %s", lt.id())
+			return nil, err
+		}
+		name := lt.id() + "["
+		out := false
+		types := []*itype{}
+		for _, c := range n.child[1:] {
+			t1, err := nodeType2(interp, sc, c, seen)
+			if err != nil {
+				return nil, err
+			}
+			if t1.cat == genericT || t1.incomplete {
+				t = lt
+				out = true
+				break
+			}
+			types = append(types, t1)
+			name += t1.id() + ","
+		}
+		if out {
+			break
+		}
+		name = strings.TrimSuffix(name, ",") + "]"
+		if sym, _, found := sc.lookup(name); found {
+			t = sym.typ
+			break
+		}
+		// A generic type is being instantiated. Generate it.
+		t, err = genType(interp, sc, name, lt, types, seen)
+
 	case interfaceType:
 		if sname := typeName(n); sname != "" {
 			if sym, _, found := sc.lookup(sname); found && sym.kind == typeSym {
@@ -998,7 +1035,7 @@ func nodeType2(interp *Interpreter, sc *scope, n *node, seen []*node) (t *itype,
 		sname := structName(n)
 		if sname != "" {
 			sym, _, found = sc.lookup(sname)
-			if found && sym.kind == typeSym {
+			if found && sym.kind == typeSym && sym.typ != nil {
 				t = structOf(sym.typ, sym.typ.field, withNode(n), withScope(sc))
 			} else {
 				t = structOf(nil, nil, withNode(n), withScope(sc))
@@ -1044,6 +1081,9 @@ func nodeType2(interp *Interpreter, sc *scope, n *node, seen []*node) (t *itype,
 		t = structOf(t, fields, withNode(n), withScope(sc))
 		t.incomplete = incomplete
 		if sname != "" {
+			if sc.sym[sname] == nil {
+				sc.sym[sname] = &symbol{index: -1, kind: typeSym, node: n}
+			}
 			sc.sym[sname].typ = t
 		}
 
@@ -1076,6 +1116,58 @@ func nodeType2(interp *Interpreter, sc *scope, n *node, seen []*node) (t *itype,
 	return t, err
 }
 
+func genType(interp *Interpreter, sc *scope, name string, lt *itype, types []*itype, seen []*node) (t *itype, err error) {
+	// A generic type is being instantiated. Generate it.
+	g, _, err := genAST(sc, lt.node.anc, types)
+	if err != nil {
+		return nil, err
+	}
+	t, err = nodeType2(interp, sc, g.lastChild(), seen)
+	if err != nil {
+		return nil, err
+	}
+	lt.instance = append(lt.instance, t)
+	// Add generated symbol in the scope of generic source and user.
+	sc.sym[name] = &symbol{index: -1, kind: typeSym, typ: t, node: g}
+	if lt.scope.sym[name] == nil {
+		lt.scope.sym[name] = sc.sym[name]
+	}
+
+	for _, nod := range lt.method {
+		if err := genMethod(interp, sc, t, nod, types); err != nil {
+			return nil, err
+		}
+	}
+	return t, err
+}
+
+func genMethod(interp *Interpreter, sc *scope, t *itype, nod *node, types []*itype) error {
+	gm, _, err := genAST(sc, nod, types)
+	if err != nil {
+		return err
+	}
+	if gm.typ, err = nodeType(interp, sc, gm.child[2]); err != nil {
+		return err
+	}
+	t.addMethod(gm)
+
+	// If the receiver is a pointer to a generic type, generate also the pointer type.
+	if rtn := gm.child[0].child[0].lastChild(); rtn != nil && rtn.kind == starExpr {
+		pt := ptrOf(t, withNode(t.node), withScope(sc))
+		pt.addMethod(gm)
+		rtn.typ = pt
+	}
+
+	// Compile the method AST in the scope of the generic type.
+	scop := nod.typ.scope
+	if _, err = interp.cfg(gm, scop, scop.pkgID, scop.pkgName); err != nil {
+		return err
+	}
+
+	// Generate closures for function body.
+	return genRun(gm)
+}
+
 // findPackageType searches the top level scope for a package type.
 func findPackageType(interp *Interpreter, sc *scope, n *node) *itype {
 	// Find the root scope, the package symbols will exist there.
@@ -1242,8 +1334,13 @@ func (t *itype) numOut() int {
 		if t.rtype.Kind() == reflect.Func {
 			return t.rtype.NumOut()
 		}
+	case builtinT:
+		switch t.name {
+		case "append", "cap", "complex", "copy", "imag", "len", "make", "new", "real", "recover":
+			return 1
+		}
 	}
-	return 1
+	return 0
 }
 
 func (t *itype) out(i int) *itype {
@@ -1266,7 +1363,7 @@ func (t *itype) concrete() *itype {
 }
 
 func (t *itype) underlying() *itype {
-	if t.cat == aliasT {
+	if t.cat == linkedT {
 		return t.val.underlying()
 	}
 	return t
@@ -1274,10 +1371,10 @@ func (t *itype) underlying() *itype {
 
 // typeDefined returns true if type t1 is defined from type t2 or t2 from t1.
 func typeDefined(t1, t2 *itype) bool {
-	if t1.cat == aliasT && t1.val == t2 {
+	if t1.cat == linkedT && t1.val == t2 {
 		return true
 	}
-	if t2.cat == aliasT && t2.val == t1 {
+	if t2.cat == linkedT && t2.val == t1 {
 		return true
 	}
 	return false
@@ -1313,7 +1410,7 @@ func isComplete(t *itype, visited map[string]bool) bool {
 		visited[name] = true
 	}
 	switch t.cat {
-	case aliasT:
+	case linkedT:
 		if t.val != nil && t.val.cat != nilT {
 			// A type aliased to a partially defined type is considered complete, to allow recursivity.
 			return true
@@ -1357,7 +1454,7 @@ func (t *itype) assignableTo(o *itype) bool {
 		return true
 	}
 
-	if t.cat == aliasT && o.cat == aliasT && (t.underlying().id() != o.underlying().id() || !typeDefined(t, o)) {
+	if t.cat == linkedT && o.cat == linkedT && (t.underlying().id() != o.underlying().id() || !typeDefined(t, o)) {
 		return false
 	}
 
@@ -1472,7 +1569,7 @@ func (t *itype) methods() methodSet {
 		seen[typ] = true
 
 		switch typ.cat {
-		case aliasT:
+		case linkedT:
 			for k, v := range getMethods(typ.val) {
 				res[k] = v
 			}
@@ -1563,7 +1660,7 @@ func (t *itype) zero() (v reflect.Value, err error) {
 		return v, err
 	}
 	switch t.cat {
-	case aliasT:
+	case linkedT:
 		v, err = t.val.zero()
 
 	case arrayT, ptrT, structT, sliceT:
@@ -1581,7 +1678,7 @@ func (t *itype) zero() (v reflect.Value, err error) {
 // fieldIndex returns the field index from name in a struct, or -1 if not found.
 func (t *itype) fieldIndex(name string) int {
 	switch t.cat {
-	case aliasT, ptrT:
+	case linkedT, ptrT:
 		return t.val.fieldIndex(name)
 	}
 	for i, field := range t.field {
@@ -1616,7 +1713,7 @@ func (t *itype) lookupField(name string) []int {
 		seen[typ] = true
 
 		switch typ.cat {
-		case aliasT, ptrT:
+		case linkedT, ptrT:
 			return lookup(typ.val)
 		}
 		if fi := typ.fieldIndex(name); fi >= 0 {
@@ -1625,7 +1722,7 @@ func (t *itype) lookupField(name string) []int {
 
 		for i, f := range typ.field {
 			switch f.typ.cat {
-			case ptrT, structT, interfaceT, aliasT:
+			case ptrT, structT, interfaceT, linkedT:
 				if index2 := lookup(f.typ); len(index2) > 0 {
 					return append([]int{i}, index2...)
 				}
@@ -1684,7 +1781,7 @@ func (t *itype) methodCallType() reflect.Type {
 }
 
 func (t *itype) resolveAlias() *itype {
-	for t.cat == aliasT {
+	for t.cat == linkedT {
 		t = t.val
 	}
 	return t
@@ -1728,7 +1825,7 @@ func (t *itype) lookupMethod2(name string, seen map[*itype]bool) (*node, []int)
 				}
 			}
 		}
-		if t.cat == aliasT || isInterfaceSrc(t) && t.val != nil {
+		if t.cat == linkedT || isInterfaceSrc(t) && t.val != nil {
 			return t.val.lookupMethod2(name, seen)
 		}
 	}
@@ -1842,6 +1939,7 @@ type refTypeContext struct {
 	// "top-level" point.
 	rect       *itype
 	rebuilding bool
+	slevel     int
 }
 
 // Clone creates a copy of the ref type context.
@@ -1945,7 +2043,7 @@ func (t *itype) refType(ctx *refTypeContext) reflect.Type {
 		return reflect.TypeOf((*generic)(nil)).Elem()
 	}
 	switch t.cat {
-	case aliasT:
+	case linkedT:
 		t.rtype = t.val.refType(ctx)
 	case arrayT:
 		t.rtype = reflect.ArrayOf(t.length, t.val.refType(ctx))
@@ -1981,11 +2079,19 @@ func (t *itype) refType(ctx *refTypeContext) reflect.Type {
 	case mapT:
 		t.rtype = reflect.MapOf(t.key.refType(ctx), t.val.refType(ctx))
 	case ptrT:
-		t.rtype = reflect.PtrTo(t.val.refType(ctx))
+		rt := t.val.refType(ctx)
+		if rt == unsafe2.DummyType && ctx.slevel > 1 {
+			// We have a pointer to a recursive struct which is not yet fully computed.
+			// Return it but do not yet store it in rtype, so the complete version can
+			// be stored in future.
+			return reflect.PtrTo(rt)
+		}
+		t.rtype = reflect.PtrTo(rt)
 	case structT:
 		if t.name != "" {
 			ctx.defined[name] = t
 		}
+		ctx.slevel++
 		var fields []reflect.StructField
 		for i, f := range t.field {
 			field := reflect.StructField{
@@ -2002,6 +2108,7 @@ func (t *itype) refType(ctx *refTypeContext) reflect.Type {
 				}
 			}
 		}
+		ctx.slevel--
 		fieldFix := []int{} // Slice of field indices to fix for recursivity.
 		t.rtype = reflect.StructOf(fields)
 		if ctx.isComplete() {
@@ -2051,14 +2158,12 @@ func (t *itype) frameType() (r reflect.Type) {
 		panic(err)
 	}
 	switch t.cat {
-	case aliasT:
+	case linkedT:
 		r = t.val.frameType()
 	case arrayT:
 		r = reflect.ArrayOf(t.length, t.val.frameType())
 	case sliceT, variadicT:
 		r = reflect.SliceOf(t.val.frameType())
-	case funcT:
-		r = reflect.TypeOf((*node)(nil))
 	case interfaceT:
 		if len(t.field) == 0 {
 			// empty interface, do not wrap it
@@ -2182,16 +2287,6 @@ func constToString(v reflect.Value) string {
 	return constant.StringVal(c)
 }
 
-func defRecvType(n *node) *itype {
-	if n.kind != funcDecl || len(n.child[0].child) == 0 {
-		return nil
-	}
-	if r := n.child[0].child[0].lastChild(); r != nil {
-		return r.typ
-	}
-	return nil
-}
-
 func wrappedType(n *node) *itype {
 	if n.typ.cat != valueT {
 		return nil
@@ -2210,7 +2305,7 @@ func isShiftNode(n *node) bool {
 // chanElement returns the channel element type.
 func chanElement(t *itype) *itype {
 	switch t.cat {
-	case aliasT:
+	case linkedT:
 		return chanElement(t.val)
 	case chanT, chanSendT, chanRecvT:
 		return t.val
@@ -2234,12 +2329,16 @@ func isGeneric(t *itype) bool {
 	return t.cat == funcT && t.node != nil && len(t.node.child) > 0 && len(t.node.child[0].child) > 0
 }
 
+func isNamedFuncSrc(t *itype) bool {
+	return isFuncSrc(t) && t.node.anc.kind == funcDecl
+}
+
 func isFuncSrc(t *itype) bool {
-	return t.cat == funcT || (t.cat == aliasT && isFuncSrc(t.val))
+	return t.cat == funcT || (t.cat == linkedT && isFuncSrc(t.val))
 }
 
 func isPtrSrc(t *itype) bool {
-	return t.cat == ptrT || (t.cat == aliasT && isPtrSrc(t.val))
+	return t.cat == ptrT || (t.cat == linkedT && isPtrSrc(t.val))
 }
 
 func isSendChan(t *itype) bool {
@@ -2256,7 +2355,7 @@ func isArray(t *itype) bool {
 }
 
 func isInterfaceSrc(t *itype) bool {
-	return t.cat == interfaceT || (t.cat == aliasT && isInterfaceSrc(t.val))
+	return t.cat == interfaceT || (t.cat == linkedT && isInterfaceSrc(t.val))
 }
 
 func isInterfaceBin(t *itype) bool {
@@ -2271,7 +2370,7 @@ func isBin(t *itype) bool {
 	switch t.cat {
 	case valueT:
 		return true
-	case aliasT, ptrT:
+	case linkedT, ptrT:
 		return isBin(t.val)
 	default:
 		return false
@@ -2284,7 +2383,7 @@ func isStruct(t *itype) bool {
 	switch t.cat {
 	case structT:
 		return true
-	case aliasT, ptrT:
+	case linkedT, ptrT:
 		return isStruct(t.val)
 	case valueT:
 		k := t.rtype.Kind()
